@@ -62,12 +62,13 @@
 # define MSG_DONTWAIT 0
 #endif
 
-struct dns_qlink {
-  struct dns_query *next, *prev;
+struct dns_qlist {
+  struct dns_query *head, *tail;
 };
 
 struct dns_query {
-  struct dns_qlink dnsq_link;		/* list entry (should be first) */
+  struct dns_query *dnsq_next;          /* double-linked list */
+  struct dns_query *dnsq_prev;
   unsigned dnsq_origdnl0;		/* original query DN len w/o last 0 */
   unsigned dnsq_flags;			/* control flags for this query */
   unsigned dnsq_servi;			/* index of next server to try */
@@ -91,60 +92,40 @@ struct dns_query {
 
 /* working with dns_query lists */
 
-static __inline void qlist_init(struct dns_qlink *list) {
-  list->next = list->prev = (struct dns_query *)list;
+static __inline void qlist_init(struct dns_qlist *list) {
+  list->head = list->tail = NULL;
 }
 
-static __inline int qlist_isempty(const struct dns_qlink *list) {
-  return list->next == (const struct dns_query *)list ? 1 : 0;
-}
-
-static __inline struct dns_query *qlist_first(struct dns_qlink *list) {
-  return list->next == (struct dns_query *)list ? 0 : list->next;
-}
-
-static __inline void qlist_remove(struct dns_query *q) {
-  q->dnsq_link.next->dnsq_link.prev = q->dnsq_link.prev;
-  q->dnsq_link.prev->dnsq_link.next = q->dnsq_link.next;
-}
-
-/* insert q between prev and next */
-static __inline void
-qlist_insert(struct dns_query *q,
-             struct dns_query *prev, struct dns_query *next) {
-  q->dnsq_link.next = next;
-  q->dnsq_link.prev = prev;
-  prev->dnsq_link.next = next->dnsq_link.prev = q;
+static __inline void qlist_remove(struct dns_qlist *list, struct dns_query *q) {
+   if (q->dnsq_prev) q->dnsq_prev->dnsq_next = q->dnsq_next;
+   else list->head = q->dnsq_next;
+   if (q->dnsq_next) q->dnsq_next->dnsq_prev = q->dnsq_prev;
+   else list->tail = q->dnsq_prev;
 }
 
 static __inline void
-qlist_insert_after(struct dns_query *q, struct dns_query *prev) {
-  qlist_insert(q, prev, prev->dnsq_link.next);
+qlist_add_head(struct dns_qlist *list, struct dns_query *q) {
+  q->dnsq_next = list->head;
+  if (list->head) list->head->dnsq_prev = q;
+  else list->tail = q;
+  list->head = q;
+  q->dnsq_prev = NULL;
 }
 
 static __inline void
-qlist_insert_before(struct dns_query *q, struct dns_query *next) {
-  qlist_insert(q, next->dnsq_link.prev, next);
+qlist_insert_after(struct dns_qlist *list,
+                   struct dns_query *q, struct dns_query *prev) {
+  if ((q->dnsq_prev = prev) != NULL) {
+    if ((q->dnsq_next = prev->dnsq_next) != NULL)
+      q->dnsq_next->dnsq_prev = q;
+    else
+      list->tail = q;
+    prev->dnsq_next = q;
+  }
+  else
+    qlist_add_head(list, q);
 }
 
-static __inline void
-qlist_add_tail(struct dns_query *q, struct dns_qlink *top) {
-  qlist_insert_before(q, (struct dns_query *)top);
-}
-
-static __inline void
-qlist_add_head(struct dns_query *q, struct dns_qlink *top) {
-  qlist_insert_after(q, (struct dns_query *)top);
-}
-
-#define QLIST_FIRST(list, direction) ((list)->direction)
-#define QLIST_ISLAST(list, q) ((q) == (struct dns_query*)(list))
-#define QLIST_NEXT(q, direction) ((q)->dnsq_link.direction)
-
-#define QLIST_FOR_EACH(list, q, direction) \
-  for(q = QLIST_FIRST(list, direction); \
-      !QLIST_ISLAST(list, q); q = QLIST_NEXT(q, direction))
-
 union sockaddr_ns {
   struct sockaddr sa;
   struct sockaddr_in sin;
@@ -185,7 +166,7 @@ struct dns_ctx {		/* resolver context */
   struct udns_jranctx dnsc_jran;	/* random number generator state */
   unsigned dnsc_nextid;			/* next queue ID to use if !0 */
   int dnsc_udpsock;			/* UDP socket */
-  struct dns_qlink dnsc_qactive;	/* active list sorted by deadline */
+  struct dns_qlist dnsc_qactive;	/* active list sorted by deadline */
   int dnsc_nactive;			/* number entries in dnsc_qactive */
   dnsc_t *dnsc_pbuf;			/* packet buffer (udpbuf size) */
   int dnsc_qstatus;			/* last query status value */
@@ -229,10 +210,12 @@ struct dns_ctx dns_defctx;
 static void dns_assert_ctx(const struct dns_ctx *ctx) {
   int nactive = 0;
   const struct dns_query *q;
-  QLIST_FOR_EACH(&ctx->dnsc_qactive, q, next) {
+  for(q = ctx->dnsc_qactive.head; q; q = q->dnsq_next) {
     assert(q->dnsq_ctx == ctx);
-    assert(q->dnsq_link.next->dnsq_link.prev == q);
-    assert(q->dnsq_link.prev->dnsq_link.next == q);
+    assert(q == (q->dnsq_next ?
+                 q->dnsq_next->dnsq_prev : ctx->dnsc_qactive.tail));
+    assert(q == (q->dnsq_prev ?
+                 q->dnsq_prev->dnsq_next : ctx->dnsc_qactive.head));
     ++nactive;
   }
   assert(nactive == ctx->dnsc_nactive);
@@ -374,7 +357,7 @@ _dns_request_utm(struct dns_ctx *ctx, time_t now) {
   struct dns_query *q;
   time_t deadline;
   int timeout;
-  q = qlist_first(&ctx->dnsc_qactive);
+  q = ctx->dnsc_qactive.head;
   if (!q)
     deadline = -1, timeout = -1;
   else if (!now || q->dnsq_deadline <= now)
@@ -433,7 +416,7 @@ static void dns_init_rng(struct dns_ctx *ctx) {
 }
 
 void dns_close(struct dns_ctx *ctx) {
-  struct dns_query *q;
+  struct dns_query *q, *p;
   SETCTX(ctx);
   if (CTXINITED(ctx)) {
     if (ctx->dnsc_udpsock >= 0)
@@ -442,10 +425,12 @@ void dns_close(struct dns_ctx *ctx) {
     if (ctx->dnsc_pbuf)
       free(ctx->dnsc_pbuf);
     ctx->dnsc_pbuf = NULL;
-    while((q = qlist_first(&ctx->dnsc_qactive)) != NULL) {
-      qlist_remove(q);
-      free(q);
+    q = ctx->dnsc_qactive.head;
+    while((p = q) != NULL) {
+      q = q->dnsq_next;
+      free(p);
     }
+    qlist_init(&ctx->dnsc_qactive);
     ctx->dnsc_nactive = 0;
     dns_drop_utm(ctx);
   }
@@ -628,7 +613,7 @@ dns_end_query(struct dns_ctx *ctx, struct dns_query *q,
   assert(cbck != 0);	/*XXX callback may be NULL */
   assert(ctx->dnsc_nactive > 0);
   --ctx->dnsc_nactive;
-  qlist_remove(q);
+  qlist_remove(&ctx->dnsc_qactive, q);
   /* force the query to be unconnected */
   /*memset(q, 0, sizeof(*q));*/
 #ifndef NDEBUG
@@ -722,10 +707,10 @@ static void dns_newid(struct dns_ctx *ctx, struct dns_query *q) {
     c0 = ctx->dnsc_nextid & 0xff;
     c1 = (ctx->dnsc_nextid >> 8) & 0xff;
     ctx->dnsc_nextid >>= 16;
-    QLIST_FOR_EACH(&ctx->dnsc_qactive, c, next)
+    for(c = ctx->dnsc_qactive.head; c; c = c->dnsq_next)
       if (c->dnsq_id[0] == c0 && c->dnsq_id[1] == c1)
         break; /* found such entry, try again */
-    if (QLIST_ISLAST(&ctx->dnsc_qactive, c))
+    if (!c)
       break;
   }
   q->dnsq_id[0] = c0; q->dnsq_id[1] = c1;
@@ -837,13 +822,13 @@ dns_send_this(struct dns_ctx *ctx, struct dns_query *q,
     (dns_find_serv(ctx, q) ? 1 : ctx->dnsc_timeout << q->dnsq_try);
 
   /* move the query to the proper place, according to the new deadline */
-  qlist_remove(q);
+  qlist_remove(&ctx->dnsc_qactive, q);
   { /* insert from the tail */
     struct dns_query *p;
-    QLIST_FOR_EACH(&ctx->dnsc_qactive, p, prev)
+    for(p = ctx->dnsc_qactive.tail; p; p = p->dnsq_prev)
       if (p->dnsq_deadline <= q->dnsq_deadline)
 	break;
-    qlist_insert_after(q, p);
+    qlist_insert_after(&ctx->dnsc_qactive, q, p);
   }
 
   return 0;
@@ -880,9 +865,10 @@ dns_send(struct dns_ctx *ctx, struct dns_query *q, time_t now) {
   dns_send_this(ctx, q, q->dnsq_servi++, now);
 }
 
-static void dns_dummy_cb(struct dns_ctx *ctx, void *result, void *data) {
+static void
+dns_dummy_cb(struct dns_ctx *unused_ctx, void *result, void *unused_data) {
   if (result) free(result);
-  data = ctx = 0;	/* used */
+  (void)unused_ctx; (void)unused_data; /* avoid warning */
 }
 
 /* The (only, main, real) query submission routine.
@@ -939,7 +925,7 @@ dns_submit_dn(struct dns_ctx *ctx,
    * to insert it into the head of the list.  Next call to dns_timeouts()
    * will actually send it.
    */
-  qlist_add_head(q, &ctx->dnsc_qactive);
+  qlist_add_head(&ctx->dnsc_qactive, q);
   ++ctx->dnsc_nactive;
   dns_request_utm(ctx, 0);
 
@@ -1028,8 +1014,8 @@ again: /* receive the reply */
   }
 
   /* find the matching query, by qID */
-  for (q = QLIST_FIRST(&ctx->dnsc_qactive, next);; q = QLIST_NEXT(q, next)) {
-    if (QLIST_ISLAST(&ctx->dnsc_qactive, q)) {
+  for (q = ctx->dnsc_qactive.head; ; q = q->dnsq_next) {
+    if (!q) {
       /* no more requests: old reply? */
       DNS_DBG(ctx, -5/*no matching query*/, &sns.sa, slen, pbuf, r);
       goto again;
@@ -1228,7 +1214,7 @@ int dns_timeouts(struct dns_ctx *ctx, int maxwait, time_t now) {
    * If not, this is the query which determines the closest deadline.
    */
 
-  q = qlist_first(&ctx->dnsc_qactive);
+  q = ctx->dnsc_qactive.head;
   if (!q)
     return maxwait;
   if (!now)
@@ -1244,7 +1230,7 @@ int dns_timeouts(struct dns_ctx *ctx, int maxwait, time_t now) {
       /* process expired deadline */
       dns_send(ctx, q, now);
     }
-  } while((q = qlist_first(&ctx->dnsc_qactive)) != NULL);
+  } while((q = ctx->dnsc_qactive.head) != NULL);
 
   dns_request_utm(ctx, now); /* update timer with new deadline */
   return maxwait;
@@ -1273,8 +1259,8 @@ void *dns_resolve(struct dns_ctx *ctx, struct dns_query *q) {
 
   assert(ctx == q->dnsq_ctx);
   dns_assert_ctx(ctx);
-  /* do not allow re-resolving syncronous queries */
-  assert(q->dnsq_cbck != dns_resolve_cb && "can't resolve syncronous query");
+  /* do not allow re-resolving synchronous queries */
+  assert(q->dnsq_cbck != dns_resolve_cb && "can't resolve synchronous query");
   if (q->dnsq_cbck == dns_resolve_cb) {
     ctx->dnsc_qstatus = DNS_E_BADQUERY;
     return NULL;
@@ -1326,11 +1312,11 @@ int dns_cancel(struct dns_ctx *ctx, struct dns_query *q) {
   SETCTX(ctx);
   dns_assert_ctx(ctx);
   assert(q->dnsq_ctx == ctx);
-  /* do not allow cancelling syncronous queries */
-  assert(q->dnsq_cbck != dns_resolve_cb && "can't cancel syncronous query");
+  /* do not allow cancelling synchronous queries */
+  assert(q->dnsq_cbck != dns_resolve_cb && "can't cancel synchronous query");
   if (q->dnsq_cbck == dns_resolve_cb)
     return (ctx->dnsc_qstatus = DNS_E_BADQUERY);
-  qlist_remove(q);
+  qlist_remove(&ctx->dnsc_qactive, q);
   --ctx->dnsc_nactive;
   dns_request_utm(ctx, 0);
   return 0;
